import java.applet.*;
import java.awt.*;
import java.util.*;

class Sky extends Canvas {
	Image offImage;
	Graphics offGraphics = null;
	StarMap map = new StarMap();
	public int draw_lines = 1;
	public int draw_stars = 1;
	public int draw_horizon = 0;
	public int draw_equator = 0;
	public int draw_animation = 0;

	public boolean write_text_on_screen = false;
	public String errmsg;


	int Width = 400;
	int Height = 300;

	Dimension d;
	
	double t, longitude, latitude;
	GregorianDate gd;
	int azimut, elevation;

	/**
	Sky constructor. Here is drawn all what is shown on the applet screen.

	@param gd <b>GregorianDate</b>
	@param longitude observer's longitude
	@param latitude observer's latitude
	@param time observer's local time
	@param azimut observer's azimut
	@param elevation observer's elevation

	@return new <b>Sky</b> object
	*/
	public Sky( GregorianDate gd, double longitude, double latitude, double time, int azimut, int elevation ) {
		this.gd = gd;
		this.longitude = longitude;
		this.latitude = latitude;
		this.t = time;
		this.elevation = elevation;
		this.azimut  = azimut;
		setBackground( new Color ( 0.0f, 0.0f, 0.5f)  );
	}


	/**
	Sets local time.

	@param t local time
	*/
	public void setTime( double t ) {
		this.t = t;
	}

	/**
	Sets observer's azimut.

	@param azimut observer's azimut
	*/
	public void setAzimut( int azimut ) {
		this.azimut = azimut;
	}


	/**
	Sets observer's elevation.

	@param elevation observer's elevation.
	*/
	public void setElevation( int elevation ) {
		this.elevation = elevation;
	}

	/**
	Sets observer's latitude.

	@param latitude observer's latitude
	*/
	public void setLatitude( double latitude ) {
		this.latitude = latitude;
	}

	/**
	Sets observer's longitude.

	@param longitude observer's longitude
	*/
	public void setLongitude( double longitude ) {
		this.longitude = longitude;
	}

	/**
	Sets <b>GregorianDate</b>.
	
	@param gd <b>GregorianDate</b>
	*/
	public void setGregorianDate( GregorianDate gd ) {
		this.gd = gd;
	}

	public void update (Graphics g) {
		paint(g);
	}


	/**
		Draws error messages on screen...
	*/
	private void drawError(Graphics g, String err, int width, int height) {
			int ypos = 0;
			int i=1;
			/* default delimiter :\t\r\n, don't want to return delimiter tokens */
			StringTokenizer st = new StringTokenizer( errmsg, "\n", false );
			Font f = getFont() ;
			int LINE_SPACE = 10;
			Color c = g.getColor();
			String str;			
			int font_height = getFontMetrics(f).getHeight(); 

			g.clearRect( 0, 0, Width, Height );
			
			
			while ( st.hasMoreTokens() ) {
				str = st.nextToken();
				g.setColor( Color.yellow );
				g.drawString("Error #"+i++ +": "+str , 10, ypos += ( font_height + LINE_SPACE ) );
			}

			g.setColor( c );
	}

	/**
		The main paint() function.
		It has two states: it draws stars and constellations or 
		it writes error in console-style screen.
	*/
	public void paint( Graphics g ) {
		d = size();
		Width = d.width;
		Height = d.height;


		map.setWidth( Width );

		/* An error has occured, so draw flush errors on screen and return immidiately */
		if ( write_text_on_screen ) {
			drawError( g, errmsg, Width, Height );
			return;
		}
		
		/* Normal drawing process, no errors */
		map.setHeight( Height );

		if ( offGraphics == null ) {
			offImage    = createImage( Width, Height );
			offGraphics = offImage.getGraphics();
		}

		offGraphics.clearRect( 0, 0, Width, Height );
		map.computeMap( offGraphics, gd, t, longitude, latitude, azimut, elevation );
		if ( draw_lines == 1) {
			map.drawLines( offGraphics );
		}
		if ( draw_stars == 1) {
			map.drawStars( offGraphics );
		}
		if ( draw_horizon == 1 ) { 
			map.drawHorizon( offGraphics, azimut, elevation );
		}
		if ( draw_equator == 1 ) {
			map.drawEquator( offGraphics, azimut, elevation, latitude, longitude );
		}

		g.drawImage( offImage, 0, 0, Width, Height, null );
		
	}


}

/* NOTE:
 * local time `t' is measured in hours, whenever it is needed it is converted into degrees.
 * longitude is also measured in hours from GMT, also automatically converted into degrees.
 * latitude is measured in degrees.
 */

public class Main extends Applet implements Runnable {

	double Y = 2002;
	double M = 3;
	double D = 13;
	double t = 6;
	double longitude = 1;
	double latitude = 50;

	final int START_AZIMUT = 180;
	final int START_ELEVATION = 0;
	int azimut = START_AZIMUT;
	int elevation = START_ELEVATION;

	int appletWidth = 640;
	int appletHeight = 480;

	Thread anime = null;

	GregorianDate gd = new GregorianDate (Y, M, D);

	Sky sky;

	String [] month_names = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Nov", "Oct", "Dec" };
	

	Scrollbar horiz_slide = new Scrollbar( Scrollbar.HORIZONTAL, START_AZIMUT, 1, 0, 360);
	Scrollbar vert_slide  = new Scrollbar( Scrollbar.VERTICAL,  START_ELEVATION, 1, -90, 90);
	Button    anime_control = new Button( "Start animation" );
	TextField year_field = new TextField( Integer.toString((int)Y), 4);	
	TextField month_field = new TextField( Integer.toString((int)M), 4);	
	TextField day_field = new TextField( Integer.toString((int)D), 4);	
	TextField hour_field = new TextField( Integer.toString((int)t), 4);	
	TextField minute_field = new TextField( Double.toString( (t-(int)t) * 60 ), 4);	
	TextField longitude_field = new TextField( Double.toString( 15 * longitude ), 4);	
	TextField latitude_field = new TextField( Double.toString( latitude ), 4);	
	Button    draw_lines = new Button( "Hide Lines" );
	Button    draw_stars = new Button( "Hide Stars" );
	Button    draw_equator = new Button( "Draw Equator" );
	Button    draw_horizon = new Button( "Draw Horizon" );

	public void init() {



		Panel north_panel = new Panel();
		Panel south_panel = new Panel();

		Panel west_panel = new Panel();


		/* Layout */
		setLayout( new BorderLayout() );
		add( "North", horiz_slide );
		add( "East",  vert_slide );
		add( "South", anime_control );

		south_panel.setLayout( new GridLayout( 7, 2 ) );
		south_panel.add( new Label("Year:") );
		south_panel.add( year_field );
		south_panel.add( new Label("Month:") );
		south_panel.add( month_field );
		south_panel.add( new Label("Day:") );
		south_panel.add( day_field );
		south_panel.add( new Label("Hour:") );
		south_panel.add( hour_field );
		south_panel.add( new Label("Minute:") );
		south_panel.add( minute_field );
		south_panel.add( new Label("Longitude:") );
		south_panel.add( longitude_field );
		south_panel.add( new Label("Latitude:") );
		south_panel.add( latitude_field );

		north_panel.setLayout( new GridLayout( 4, 1 ) );
		north_panel.add( draw_lines );
		north_panel.add( draw_stars );
		north_panel.add( draw_equator );
		north_panel.add( draw_horizon );

		west_panel.setLayout( new GridLayout ( 2, 1 ) );
		west_panel.add( north_panel );
		west_panel.add( south_panel );


		add( "West", west_panel );

		add( "Center" , sky = new Sky(gd, longitude, latitude, t, azimut, elevation ) );

		Trans3D.initTables();

	}

	/* UGH ! */
	private boolean ___convert_boolean( int val ) {
		if ( val == 0 ) return false; else return true;
	}

	/**
		Checks for the leap year.
		@return true when it is a leap year, otherwise false
	*/
	private boolean is_leap_year( int Y ) {
		if  ( ( ( (int)Y & 3 ) == 0 ) && (  ___convert_boolean((int)Y % 100) ==  ___convert_boolean((int)Y % 400) ) )  {
			/* leap year */
			return true;
		} else {
			return false;
		}
	}

	/** 
		Returns days in given month with care about year.
		@return -1 if month is out of 1-12. Otherwise it returns 28-31 depending on the given year and month.
	*/
	private int days_per_month( int M, int Y ) {
		int days_in_feb;

		if ( M < 1 || M > 12 ) return -1;

		/* February at first ... */
		if ( M==2 ) {
			/* leap year */
			if ( is_leap_year( Y ) ) 
				return 29;
			else 
				return 28;
		}

		if ( M<8 ) {
			return 30 + (M&1);
		} else {
			return 31 - (M&1);	
		}		
	}

	/** 
		Checks whether the input values are correct.

		@return  true if input is correct, false otherwise.
	*/
	public boolean checkInput() {
		String err = new String(" ");
		double hh,mm;
		int days;
		boolean rc = true;



		hh = Double.valueOf( hour_field.getText() ).doubleValue() ;
		mm = Double.valueOf( minute_field.getText() ).doubleValue() ;


		/* month check */
		if ( M<1 || M>12 ) {
			err += new String("Month could be between 1-12!\n");
		}

		if ( D<1 ) {
			err += new String("Day number must be positive!\n");
		}

		days = days_per_month( (int) M, (int) Y );
		
		if ( days == -1 ) {
			/* some error informing about bad month could be here in general, 
			   but in this code it is above. */
		}
		else
		if ( D > days ) {
			err += new String("This month has only "+days+" days.\n");
		}

		if ( hh > 23 || hh < 0 ) {
			err += new String("Hours must be between 0-23!\n");
		}

		
		if ( mm >= 60 || mm < 0 ) {
			err += new String("Minutes must be between <0-60)!\n");
		}

		if ( err.length() > 1 ) {
			writeError( err );
			rc = false;
		}
	
		return rc;		
	}


	public void start() {
		/* Restart the animation only if it was played before */
		if ( sky.draw_animation == 1 ) {
			start_anime(); 
		}
	}

	/**
	This is start of the <b>Sky</b> rotating animation.
	*/
	public void start_anime() {
		if ( anime == null ) {
			anime = new Thread(this);
			anime.start();
		}
	}

	/**
	Sometimes we need to stop the animation.
	*/
	public void stop_anime() {
		anime = null;
	}

	public void stop() {
		stop_anime();
	}

	public void run() {

		boolean change_day, change_month;

		while ( anime != null ) {
			change_month = change_day = false;

			t += 0.2;
			if ( t > 24 ) {
				D++;
				change_day = true;

				if ( D > days_per_month( (int)M, (int) Y ) ) {
					change_month = true;
					D = 1;
					M++;

					if ( M > 12 ) {
						M = 1;
						Y++;
						year_field.setText( Double.toString( Y ) );
					}
				}

				if ( change_day )     day_field.setText( Integer.toString( (int) D ) );
				if ( change_month )   month_field.setText( Integer.toString( ( int ) M ) );

				gd.setY( Y );
				gd.setM( M );
				gd.setD( D );
				sky.setGregorianDate( gd );
				t = 0;
			}

			sky.setTime( t );
			
			hour_field.setText( Integer.toString( (int) Math.floor( t ) ) );
			minute_field.setText( Double.toString( ( t - Math.floor( t ) ) * 60 ) ) ;

			sky.repaint();

			try {
				Thread.sleep (20);
			}

			catch ( InterruptedException e ) {};
		}
	}

	public void writeError( String errmsg ) {
		stop_anime();
		anime_control.setLabel( "Start animation" );
		if (sky.draw_animation == 1 ) sky.draw_animation ^= 1;
		sky.write_text_on_screen = true;
		sky.errmsg = errmsg;
		sky.repaint();
	}


	/**
	Button and Textfiled are handled here.
	*/
	public boolean action ( Event e, Object arg ) {
		int month;

		/* Draw Stars button */
		if ( e.target == draw_stars ) {
			sky.draw_stars ^= 1;
			if ( sky.draw_stars == 1 ) {
				draw_stars.setLabel( "Hide stars" );
			} else {
				draw_stars.setLabel( "Draw stars" );
			}
			/* If animation is not running, we must repaint the window manually */
			if ( anime == null ) {
				sky.repaint();
			}

			return true;
		}

		/* Draw Lines button */
		if ( e.target == draw_lines ) {
			sky.draw_lines ^= 1;
			if ( sky.draw_lines == 1 ) {
				draw_lines.setLabel( "Hide lines" );
			} else {
				draw_lines.setLabel( "Draw lines" );
			}

			/* If animation is not running, we must repaint the window manually */
			if ( anime == null ) {
				sky.repaint();
			}

			return true;
		}

		/* Draw Horizon button */
		if ( e.target == draw_horizon ) {
			sky.draw_horizon ^= 1;
			if ( sky.draw_horizon == 1 ) {
				draw_horizon.setLabel( "Hide horizon" );
			} else {
				draw_horizon.setLabel( "Draw horizon" );
			}

			/* If animation is not running, we must repaint the window manually */
			if ( anime == null ) {
				sky.repaint();
			}

			return true;
		}

		/* Draw Equator button */
		if ( e.target == draw_equator ) {
			sky.draw_equator ^= 1;
			if ( sky.draw_equator == 1 ) {
				draw_equator.setLabel( "Hide equator" );
			} else {
				draw_equator.setLabel( "Draw equator" );
			}

			/* If animation is not running, we must repaint the window manually */
			if ( anime == null ) {
				sky.repaint();
			}

			return true;
		}

		/* Anime control button */
		if ( e.target == anime_control && sky.write_text_on_screen == false) {
			sky.draw_animation ^= 1;
			if ( sky.draw_animation == 1 ) {
				anime_control.setLabel( "Stop animation" );
				start_anime();
			} else {
				anime_control.setLabel( "Start animation" );
				stop_anime();
			}
			return true;
		}


		/* Year textfield */
		if ( e.target == year_field ) {
			Y = Double.valueOf( year_field.getText() ).doubleValue();

			if ( ! checkInput() ) {
				return true;
			}

			sky.write_text_on_screen = false;
			gd.setY( Y );
			sky.setGregorianDate( gd );
			sky.repaint();
			return true;
		}

		/* Month textfield */
		if ( e.target == month_field ) {
			M = Double.valueOf( month_field.getText() ).doubleValue();

			if ( ! checkInput() ) {
				return true;
			}

			sky.write_text_on_screen = false;
			gd.setM( M );
			sky.setGregorianDate( gd );
			sky.repaint();
			return true;
		}

		/* Day textfield */
		if ( e.target == day_field ) {
			D = Double.valueOf( day_field.getText() ).doubleValue();

			if ( ! checkInput() ) {
				return true;
			}

			sky.write_text_on_screen = false;
			gd.setD( D );
			sky.setGregorianDate( gd );
			sky.repaint();
			return true;
		}


		/* Hour or Minute textfield */
		if ( e.target == hour_field || e.target == minute_field ) {
			t = Double.valueOf( hour_field.getText() ).doubleValue() + 
				Double.valueOf( minute_field.getText() ).doubleValue() / 60;

			if ( ! checkInput() ) {
				return true;
			}

			sky.write_text_on_screen = false;
			sky.setTime( t );
			sky.repaint();
			return true;
		}

		/* Longitude textfield */
		if ( e.target == longitude_field ) {
			longitude = Double.valueOf( longitude_field.getText() ).doubleValue() / 15;

			if ( ! checkInput() ) {
				return true;
			}

			sky.write_text_on_screen = false;
			sky.setLongitude( longitude );
			sky.repaint();
			return true;
		}

		/* Latitude textfield */
		if ( e.target == latitude_field ) {
			latitude = Double.valueOf( latitude_field.getText() ).doubleValue();

			if ( ! checkInput() ) {
				return true;
			}

			sky.write_text_on_screen = false;
			sky.setLatitude( latitude );
			sky.repaint();
			return true;
		}

		return false;
	}
	

	/**
	Here we handle scrollbars events.
	*/
	public boolean handleEvent ( Event e ) {
		/* Horizontal slide control */
		if ( e.target == horiz_slide ) {
			azimut = horiz_slide.getValue();	
			sky.setAzimut( azimut );
			sky.repaint();
		}

		if ( e.target == vert_slide ) {
			elevation = vert_slide.getValue();
			sky.setElevation( elevation );
			sky.repaint();
		}

		return super.handleEvent(e);
	}

}
