/**
	This class is responsible for computing and drawing stars, horizon, equator,
	etc.

	In major consists of:

	<ul>
	<li>drawHorizon()
	<li>drawLine()
	<li>drawEquator()
	<li>drawStars()
	</ul>

	@author Michal Conos
	@version 0.1.1
*/

import java.awt.*;

public class StarMap {
	private int width  = 640;
	private int height = 480;

	private final int RADIUS = 450;


	private double starmap[] = {

 /* Ursa Minor, Ursae Minoris, UMi */
         2.07, 89.09, 2.02,    /* 1 ,    0       */
        17.40, 86.36, 4.36,    /* 23,    1       */  
        16.48, 82.05, 4.23,    /* 22,    2       */
        15.44, 77.52, 4.32,    /* 16,    3       */
        14.50, 74.15, 2.08,    /* 7      4       */   
        15.20, 71.55, 3.05,    /* 13     5       */
 /* Virgo, Virginis , Vir */
        13.00, 11.06, 2.84,    /* 47     6       */
        12.54,  3.32, 3.38,    /* 43     7       */
        13.33, -0.28, 3.38,    /* 79     8       */ 
        12.00,  8.40, 4.04,    /* ??     9       */
        12.40, -1.19, 2.74,    /* 29    10       */ 
        12.18, -0.32, 3.90,    /* 15    11       */
        11.49,  1.54, 3.60,    /* 5     12       */  
        13.08, -5.24, 4.38,    /* 51    13       */ 
        13.23,-11.02, 0.97,    /* 67    14       */
 /* Ursa Maior, Ursae Maioris, UMa */
        13.46, 49.26, 1.86,    /* 85    15       */
        13.22, 55.03, 2.06,    /* 79AB  16       */ 
        12.52, 56.06, 1.77,    /* 77    17       */
        12.14, 57.10, 3.31,    /* 69    18       */
        11.02, 61.53, 1.79,    /* 50    19       */
        11.00, 56.31, 2.37,    /* 48    20       */
        11.52, 53.50, 2.44,    /* 64    21       */ 
 /* Orion, Orionis, Ori */ 
         5.46, -9.41, 2.05,    /* 53    22       */
         5.39, -1.57, 1.77,    /* 50    23       */ 
         5.53,  7.24, 0.42,    /* 58    24       */
         5.34,  9.55, 3.39,    /* 39    25       */
         5.23,  6.20, 1.46,    /* 24    26       */ 
         5.30, -0.19, 2.24,    /* 34    27       */  
         5.13, -8.14, 0.13,    /* 19    28       */
         5.34, -1.13, 1.69,    /* 46    29       */
 /* Taurus, Tauri, Tau */
         4.34, 16.28, 0.86,    /* 87    30       */
         4.27, 19.08, 3.54,    /* 74    31       */
         4.18, 15.34, 3.65,    /* 54    32       */
         5.24, 28.35, 1.65,    /* 112   33       */
         5.36, 21.08, 3.03,    /* 123   34       */
         3.25,  9.35, 3.75,    /* 2     35       */
         3.47, 24.04, 5.0,     /* BU    36       */
 /* Lyra, Lyrae, Lyr */
        18.36, 38.46, 0.03,    /* 3     37       */
        18.43, 37.35, 4.36,    /* 6     38       */
        18.49, 33.20, 3.42,    /* 10    39       */
        18.58, 32.39, 3.24,    /* 14    40       */
        18.54, 36.52, 6.2,     /* d2    41       */  
 /* Cygnus, Cygni, Cyg */
        20.40, 45.11, 1.25,    /* 50    42       */
        20.21, 40.11, 2.23,    /* 37    43       */ 
        19.44, 45.04, 2.87,    /* 18    44       */
        19.16, 53.19, 3.76,    /* 1     45       */
        20.45, 33.53, 2.46,    /* 53    46       */
        21.11, 30.07, 3.20,    /* 64    47       */
        19.29, 27.54, 3.08,    /* 6     48       */  
 /* Aquila, Aquilae, Aql */
        19.49,  8.48, 0.76,    /* 53    49       */
        19.24,  3.04, 3.36,    /* 30    50       */    
        19.04, -4.55, 3.43,    /* 16    51       */
        19.04, 13.50, 2.99,    /* 17    52       */
        20.10,  0.54, 3.24,    /* 65    53       */
 /* Cassiopeia, Cassiopeiae, Cas */
         1.52, 63.33, 3.38,    /* 45    54       */
         1.24, 60.06, 2.68,    /* 37    55       */         
         0.55, 60.35, 2.39,    /* 27    56       */
         0.39, 56.24, 2.23,    /* 18    57       */
         0.07, 59.01, 2.37,    /* 11    58       */
 /* Andromeda, Andromedae, And */
         2.02, 42.13, 2.10,    /* 57    59       */
         1.08, 35.29, 2.05,    /* 43    60       */
         0.07, 28.57, 2.06,    /* 21    61       */
         0.55, 38.22, 3.87,    /* 37    62       */  
         0.48, 40.57, 4.40,    /* 35    63       */
 /* Hercules, Herculis, Her */
        17.14, 36.50, 3.16,    /* 67    64       */
        16.42, 38.58, 3.50,    /* 44    65       */
        16.40, 31.39, 2.81,    /* 40    66       */
        16.59, 30.58, 3.92,    /* 58    67       */  
        17.55, 37.15, 3.87,    /* 91    68       */  
        17.38, 46.01, 3.80,    /* 85    69       */  
        16.19, 46.22, 3.90,    /* 22    70       */
        16.08, 45.00, 4.27,    /* 11    71       */
        16.29, 21.33, 2.74,    /* 27    72       */
        17.13, 14.25, 3.06,    /* 64    73       */
        16.20, 19.13, 3.76,    /* 20    74       */ 
        17.14, 24.52, 3.13,    /* 65    75       */
        17.29, 26.08, 4.41,    /* 76    76       */
        17.45, 27.44, 3.42,    /* 86    77       */
        17.57, 30.12, 4.41,    /* 94    78       */ 
 /* Leo, Leonis, Leo */
         9.44, 23.53, 2.98,    /* 17    79       */
         9.51, 26.08, 3.88,    /* 24    80       */
        10.15, 23.33, 3.44,    /* 36    81       */
        10.18, 19.58, 1.98,    /* 41    82       */
        10.07, 12.05, 1.35,    /* 32    83       */
        10.06, 16.53, 3.53,    /* 30    84       */
        11.12, 20.40, 2.56,    /* 68    85       */  
        11.47, 14.43, 2.14,    /* 94    86       */ 
        11.12, 15.34, 3.35,    /* 70    87       */
 /* Cepheus, Cephei, Cep */
        23.38, 77.30, 3.21,    /* 35    88       */
        22.48, 66.04, 3.53,    /* 32    89       */ 
        22.28, 58.17, 4.34,    /* 27    90       */
        21.18, 62.29, 2.45,    /* 5     91       */
        21.28, 70.27, 3.23,    /* 8     92       */
        20.44, 61.44, 3.43,    /* 3     93       */ 
        20.29, 62.55, 4.28,    /* 2     94       */
        22.10, 58.05, 3.35,    /* 21    95       */
 /* Pegasus, Pegasi, Peg */
        21.30,  9.46, 2.39,    /* 8     96       */
        22.08,  6.04, 3.52,    /* 26    97       */
        22.40, 10.42, 3.40,    /* 42    98       */
        23.03, 15.04, 2.48,    /* 54    99       */   
         0.11, 15.03, 2.84,    /* 88   100       */
        23.02, 27.56, 2.50,    /*     101       */
        22.41, 30.05, 2.95,    /* 44   102       */
        22.48, 24.28, 3.48,    /* 48   103       */
 /* Bootes, Bootis, Boo */
        14.14, 19.19,-0.05,    /* 16   104       */
        14.30, 30.29, 3.59,    /* 25   105       */ 
        14.31, 38.25, 3.02,    /* 27   106       */
        15.01, 40.29, 3.50,    /* 42   107       */
        15.14, 33.24, 3.49,    /* 49   108       */
        14.44, 27.11, 2.37,    /* 36   109       */
        14.40, 13.49, 3.78,    /* 30   110       */
 /* Corona Borealis, Coronae Borealis, CrB */
        15.56, 26.57, 4.15,    /* 13   111       */
        15.42, 26.22, 3.85,    /* 8    112       */ 
        15.33, 26.48, 2.24,    /* 5    113       */
        15.26, 29.11, 3.68,    /* 3    114       */
        15.31, 31.27, 4.13,    /* 4    115       */
 /* Crater, Crateri, Crt */
        11.35, -9.8,  5.00,    /* ??   116       */
        11.25,-10.05, 4.50,    /* ??   117       */
        11.20,-14.39, 3.56,    /* 12   118       */
        11.23,-17.33, 4.08,    /* 15   119       */
        11.38,-18.00, 4.50,    /* ??   120       */
        11.55,-17.30, 4.50,    /* ??   121       */
        10.58,-18.10, 4.07,    /* 7    122       */
        11.10,-22.41, 4.48,    /* 11   123       */
 /* Libra, Librae, Lib */
        15.02,-25.11, 3.27,    /* 20   124       */
        14.49,-15.56, 2.75,    /* 9    125       */ 
        15.15, -9.18, 2.61,    /* 27   126       */
        15.34,-14.42, 3.91,    /* 38   127       */
 /* Pisces, Piscium, Psc */
         1.10, 29.57, 4.51,    /* 83   128       */
         1.30, 15.13, 3.62,    /* 99   129       */
         1.44,  9.02, 4.26,    /* 110  130       */
         2.01,  2.38, 3.82,    /* 113  131       */
         1.40,  5.22, 4.44,    /* 106  132       */ 
         1.01,  7.45, 4.28,    /* 71   133       */
         0.47,  7.27, 4.44,    /* 63   134       */
        23.58,  6.44, 4.01,    /* 28   135       */
        23.38,  5.29, 4.13,    /* 17   136       */
        23.26,  6.14, 4.28,    /* 10   137       */
        23.15,  3.09, 3.69,    /* 6    138       */
        23.26,  0.57, 5.90,    /* ??   139       */   
        23.02,  3.41, 4.52,    /* 4    140       */
 /* Aquarius, Aquarii, Aqr */
        23.13, -6.11, 4.22,    /* 90   141       */
        23.14, -9.13, 4.25,    /* 91   142       */       
        23.41,-14.41, 4.51,    /* 105  143       */
        23.21,-20.14, 3.98,    /* 98   144       */
        23.08,-21.19, 3.64,    /* 88   145       */
        22.53,-15.57, 3.28,    /* 76   146       */
        22.48,-13.44, 3.98,    /* 71   147       */
        22.51, -7.43, 3.79,    /* 73   148       */
        22.15, -7.55, 4.15,    /* 43   149       */ 
        22.05,-14.00, 4.25,    /* 33   150       */
        22.20, -1.31, 3.84,    /* 48   151       */
        22.04, -0.27, 2.93,    /* 34   152       */
        21.30, -5.41, 2.87,    /* 22   153       */
        21.08,-11.28, 4.52,    /* 13   154       */
        20.46, -9.35, 3.77,    /* 2    155       */
 /* Gemini, Geminorum, Gem */
         6.36, 16.25, 1.92,    /* 24   156       */
         6.21, 22.32, 2.87,    /* 13   157       */        
         6.42, 25.09, 2.98,    /* 27   158       */
         7.09, 30.18, 4.40,    /* 46   159       */
         7.33, 31.57, 1.58,    /* 66   160       */
         7.43, 28.05, 1.14,    /* 78   161       */
         7.42, 24.28, 3.57,    /* 77   162       */
         7.18, 22.02, 3.53,    /* 55   163       */
 /* Scorpius, Scorpii, Sco */
        16.50,-38.00, 3.03,    /* u1   164       */
        16.48,-34.15, 2.29,    /* 26   165       */
        16.34,-28.10, 2.81,    /* 23   166       */
        16.27,-26.23, 0.91,    /* 12   167       */
        16.19,-25.32, 2.88,    /* 20   168       */
        16.06,-20.36, 3.97,    /* 9    169       */
        15.58,-22.30, 2.32,    /* 7    170       */
        15.57,-26.03, 2.91    /* 6    171       */
 }; 

	private int coords[] = new int[ 2*starmap.length ];

	private int counter;
	private int cnum = 0;
	
	public int LINES  = 157;

	private int lines[]=
{
   0,  1,  1,  2,  2,  3,  3,  4,  4,  5,  6,  7,  7,  8, 30, 32,
   7,  9,  7, 10, 10, 11, 11, 12, 10, 13, 13, 14, 15, 16, 16, 17, 17, 18,
  18, 19, 19, 20, 20, 21, 21, 18, 22, 23, 23, 24, 24, 25, 25, 26, 26, 27,
  27, 28, 30, 31, 31, 32, 31, 33, 30, 34, 32, 35, 31, 36, 37, 38, 38, 39,
  39, 40, 40, 41, 42, 43, 43, 44, 44, 45, 43, 46, 46, 47, 43, 48, 49, 50,
  50, 51, 50, 52, 50, 53, 54, 55, 55, 56, 56, 57, 57, 58, 59, 60, 60, 61,
  60, 62, 62, 63, 64, 65, 65, 66, 66, 67, 67, 64, 64, 68, 68, 69, 65, 70,
  70, 71, 66, 72, 72, 73, 72, 74, 67, 75, 75, 76, 76, 77, 77, 78, 79, 80,
  80, 81, 81, 82, 82, 83, 82, 84, 82, 85, 85, 86, 85, 87, 88, 89, 89, 90,
  90, 91, 91, 92, 91, 93, 93, 94, 90, 95, 96, 97, 97, 98, 98, 99, 99,100,
 100, 61, 61,101,101, 99,101,102,101,103,104,105,105,106,106,107,107,108,  
 108,109,109,110,111,112,112,113,113,114,114,115,116,117,117,118,118,119,
 119,120,120,121,118,122,119,123,124,125,125,126,126,127,128,129,129,130,
 130,131,131,132,132,133,133,134,134,135,135,136,136,137,137,138,138,139,
 139,136,138,140,141,142,142,143,143,144,144,145,145,146,146,147,147,148,
 148,149,149,150,151,152,152,153,153,154,154,155,149,152,156,157,157,158,
 158,159,159,160,160,161,161,162,162,163,163,156,164,165,165,166,166,167,
 167,168,168,169,168,170,168,171,0,1
};

	/** 
	Constructs a <b>StarMap</b> object. Here we've coordinates of stars
	we draw.

	@return new <b>StarMap</b> object
	*/
	public StarMap() {
		counter = 0;
		cnum    = 0;
	}

	/**
	Constructs a <b>StarMap</b> with specified width and height.

	@param w width
	@param h height

	@return new <b>StarMap</b> object
	*/
	public StarMap(int w, int h) {
		counter = 0;
		cnum    = 0;
		width   = w;
		height  = h;
	}

	/**
	Sets new width for a <b>StarMap</b>

	@param width new width
	*/
	public void setWidth( int width ) {
		this.width = width;
	}

	/**
	Sets new height for a <b>StarMap</b>

	@param height new height
	*/
	public void setHeight( int height ) {
		this.height = height;
	}

	/**
	Computes coordinates of all stars and stores them. To do this we need to know observer's local time, longitude, latitude, azimut and eleveation.
	Also <b>GregorianDate</b> is needed. 

	@param g <b>Graphics</b> object, where we've drawn (currently we do not any drawing here).
	@param gd <b>GregorianDate</b>
	@param t observer's local time
	@param longitude observer's longitude
	@param latitude observer's latitude
	@param azimut observer's azimut
	@param elevation observer's elevation
	*/
	public void computeMap (Graphics g, GregorianDate gd, double t, double longitude, double latitude, int azimut, int elevation ) {
		Star star;
		
		StarTime st;

		Systemdl dl;
		SystemDLe dle = new SystemDLe (0,0,0);

		double a, d;
		double JD;	

		Point3D p3d;
		Point2D p2d;

	
		double local_time;
		
		int x,y;
		int w2 = width >> 1;
		int h2 = height >> 1;

		JD = BasicAstroTrans.convertGregorian2Julian ( gd );
		


		while ( ( star = getNextStar() ) != null )  {
			a = star.geta();
			d = star.getd();

			//System.out.println("a="+a);
			//System.out.println("d="+d);

			a *= Math.PI/180;
			d *= Math.PI/180;
 
			//System.out.println("*a="+a);
			//System.out.println("*d="+d);
			dle.setD (d);
			st = BasicAstroTrans.localStarTime ( JD, longitude, t );
			local_time = -a + st.gets() * 15 * Math.PI / 180;
//			System.out.println("mistni hvezdny cas:"+st.gets()+ " cas t= "+ t +" T'  v rad:"+local_time);
			dle.setL (Math.PI/2 - local_time);
			dle.sete ( (latitude - 90) * Math.PI / 180);
			dl = BasicAstroTrans.transformDLe2dl (dle);
/*
			System.out.println("dl.d="+dl.getd());
			System.out.println("dl.l="+dl.getl());

			System.out.println("***dl.l="+dl.getl());
*/
			
			p3d = Trans3D.transformPolartoCartesian (dl, RADIUS);
			Trans3D.rotate3DXZ( p3d, azimut );
			Trans3D.rotate3DYZ( p3d, ( elevation < 0 ) ? 360 + elevation : elevation );
			//p3d.setz (-p3d.getz());
			//p3d.setx (-p3d.getx());
			p2d = Trans3D.transformPoint3Dto2D (p3d, width >> 1);
		
			x = (int)Math.round(p2d.getx()+w2);
			y = (int)Math.round(h2 - p2d.gety());

			coords[ cnum++ ] = x;
			coords[ cnum++ ] = y;

		}


		clearCounter();
	}

	/**
	On demand draws horizon as viewed from observer's position.

	@param g where to draw
	@param azimut observer's azimut
	@param elevation observer's elevation
	*/
	public void drawHorizon( Graphics g, int azimut, int elevation ) {
		
		Point3D p3d = new Point3D( 0.0, 0.0, 0.0 );
		Point3D p3d2 = new Point3D( 0.0, 0.0, 0.0 );
		Point2D p2d;
		Point2D p2d2;
		double a = 0;
		double PI2 = 2*Math.PI;
		double rad_elevation = elevation * Math.PI / 180;
		int x, y, ox = 0, oy = 0, x2, y2;
		int w2 = width >> 1;
		int h2 = height >> 1;
		double da = Math.PI / 180;
		int az = azimut;
		int el = elevation;
		
		g.setColor( Color.white );
	
		g.drawLine(w2, 0, w2, height);

		while ( a < PI2 ) {
			p3d.setx( RADIUS * Math.sin(a) );
			p3d.sety( 0.0 );
			p3d.setz( RADIUS * Math.cos(a) );


			p3d2.setx( 0.0 );
			p3d2.sety( p3d.getx() );
			p3d2.setz( p3d.getz() );

			Trans3D.rotate3DYZ( p3d, elevation * Math.PI/180 );
			Trans3D.rotate3DYZ( p3d2, elevation * Math.PI/180 );

			p2d = Trans3D.transformPoint3Dto2D (p3d, width >> 1);
			p2d2 = Trans3D.transformPoint3Dto2D (p3d2, width >> 1);
	
			
			x = (int)Math.round(p2d.getx()+w2);
			y = (int)Math.round(h2 - p2d.gety());

			x2 = (int)Math.round(p2d2.getx()+w2);
			y2 = (int)Math.round(h2 - p2d2.gety());

			if ( a > 0 && x >= 0 && x < width && y >= 0 && y < height && ox >= 0 && ox < width && oy >=0 && oy < height) {
				g.drawLine( ox, oy, x, y );
			}

			if ( ( az % 45 )  == 0 ) {
				g.drawString(Integer.toString( az ), x, y - 7 );
				g.drawLine(x,y-5,x,y+5);
			}
			
			if ( ( el % 10 ) == 0 ) {
				g.drawString(Integer.toString( el ), x2+7, y2 );
				g.drawLine(x2-5,y2,x2+5,y2);
				
			}

			ox = x;
			oy = y;
			a += da;
			az++;
			el++;
			az %= 360;
			el %= 360;
		}
			
	}

	/**
	On demand draws equator as viewed from observer's position.

	@param g where to draw
	@param azimut observer's azimut
	@param elevation observer's elevation
	@param latitude observer's latitude
	@param longitude observer's longitude
	*/
	public void drawEquator( Graphics g, int azimut, int elevation, double latitude, double longitude ) {
		Point3D p3d = new Point3D( 0.0, 0.0, 0.0 );
		Point2D p2d;

		double a = 0;
		double da = Math.PI / 180;
		double PI2 = 2 * Math.PI;

		int x, y, ox = 0, oy = 0;
		int h2 = height >> 1;
		int w2 = width  >> 1;
		int eq = (int) ( longitude * 15 );

		g.setColor( Color.green );

		while ( a < PI2 ) {	
			p3d.setx( -RADIUS * Math.sin(a) );
			p3d.sety( 0.0 );
			p3d.setz( -RADIUS * Math.cos(a) );

			Trans3D.rotate3DYZ( p3d,  ( latitude - 90 ) * Math.PI / 180 );
			Trans3D.rotate3DXZ( p3d,  azimut * Math.PI / 180 );
			Trans3D.rotate3DYZ( p3d,  elevation * Math.PI / 180 );

			p2d = Trans3D.transformPoint3Dto2D (p3d, width >> 1);

			x = (int)Math.round(p2d.getx()+w2);
			y = (int)Math.round(h2 - p2d.gety());

			if ( a > 0 && x >= 0 && x < width && y >= 0 && y < height && ox >= 0 && ox < width && oy >=0 && oy < height) {
				g.drawLine( ox, oy, x, y );
			}
			
			
			if ( ( eq % 15 )  == 0 ) {
				g.drawString(Integer.toString( eq/15 )+"h", x, y - 7 );
				g.drawLine(x,y-5,x,y+5);
			}

			ox = x;
			oy = y;

			a += da;
			eq++;
			eq %= 360;
		}
		
	}

	/**
	On demand draws lines between stars as viewed from observer's position.

	@param g where to draw
	*/
	public void drawLines( Graphics g ) {
		int i = 0;
		int x1, x2, y1, y2;
		Star star;

		clearCounter();

		g.setColor( Color.blue );
		for ( i = 0; i < lines.length; i+=2 ) {
     		x1 = coords[ 2*lines[i]       ];
			y1 = coords[ 2*lines[i]   + 1 ];
			x2 = coords[ 2*lines[i+1]     ];
			y2 = coords[ 2*lines[i+1] + 1 ];

			if ( x1 >= 0 && x1 < width && y1 >= 0 && y1 < width &&
			     x2 >= 0 && x2 < width && y2 >= 0 && y2 < width    ) {

				g.drawLine( x1, y1, x2, y2 );
			}
		}

		clearCounter();
	}

	/**
	On demand draws stars as viewed from observer's position.

	@param g where to draw
	*/
	public void drawStars( Graphics g ) {
		int i = 0;
		int x, y;
		Star star;
		
		clearCounter();

		g.setColor( Color.yellow );
		while ( ( star = getNextStar() ) != null )  {

			x = coords[i];
			y = coords[i+1];

			if ( x >= 0 && x < width && y >=0 && y < height ) {
				drawStar (g, x, y);
			}

			i += 2;
		}

		clearCounter();
	}

	/**
	Draws one star. How it is drawn depends on its magnitude. This function is not finished
	yet, some improvements should occur.

	@param g where to draw
	@param x stars' x-position on screen
	@param y stars' y-position on screen
	*/
	public void drawStar(Graphics g, int x, int y) {
		/* star magnitude */
		double m = (6 - starmap [counter - 1])/3;
		int r;
		m++;
		r = (int) Math.floor ( m );
		if ( r < 0 ) r=3;

		switch (r) {
			case 0:
			case 1:
				g.drawLine (x, y, x+1,y);
				break;
			case 2:
				g.drawLine (x, y+1, x, y-1);
				g.drawLine (x-1, y, x+1, y);
				break;
			default:
				g.fillOval (x,y, r, r);
		}
	}

	/**
	This method is currently used private. It returns next <b>Star</b> in <b>StarMap</b>.
	
	@return next <b>Star</b> in map.
	*/
	public Star getNextStar () {
		double RA, RA_frac;
		int RA_int;

		if ( counter <  starmap.length) {
			counter += 3;
			RA = starmap[counter-3];
			RA_int = (int) Math.floor(RA);
			RA_frac = RA - RA_int;
			RA = RA_int + RA_frac/0.6;
			RA *= 15;

			return new Star (RA, starmap[counter-2]);
		} else {
			return null;
		}
	}

	/**
	Sometimes we need to clear the next star counter.
	*/
	public void clearCounter() {
		counter = 0;
		cnum    = 0;
	}
	
}
